@@ -1,41 +1,38 @@
 # -*- coding: utf-8 -*-
-import codecs
+import io
 import json
 import os
 import re
 import sys
-import shlex
 import base64
 import hashlib
-
 import contoml
-import delegator
+from first import first
 import pipfile
 import pipfile.api
+import six
 import toml
+import json as simplejson
 
-try:
-    import pathlib
-except ImportError:
-    import pathlib2 as pathlib
+from ._compat import Path
 
-from pip9 import ConfigOptionParser
 from .cmdparse import Script
 from .utils import (
+    atomic_open_for_write,
     mkdir_p,
-    convert_deps_from_pip,
     pep423_name,
     proper_case,
     find_requirements,
     is_editable,
-    is_file,
     is_vcs,
     cleanup_toml,
     is_installable_file,
     is_valid_url,
     normalize_drive,
     python_version,
-    escape_grouped_arguments,
+    safe_expandvars,
+    is_star,
+    get_workon_home,
 )
 from .environments import (
     PIPENV_MAX_DEPTH,
@@ -44,13 +41,32 @@ from .environments import (
     PIPENV_VIRTUALENV,
     PIPENV_TEST_INDEX,
     PIPENV_PYTHON,
+    PIPENV_DEFAULT_PYTHON_VERSION,
 )
 
 
 def _normalized(p):
     if p is None:
         return None
-    return normalize_drive(str(pathlib.Path(p).resolve()))
+    loc = Path(p)
+    if loc.is_absolute():
+        return normalize_drive(str(loc))
+    else:
+        try:
+            loc = loc.resolve()
+        except OSError:
+            loc = loc.absolute()
+        return normalize_drive(str(loc))
+
+
+DEFAULT_NEWLINES = u'\n'
+
+
+def preferred_newlines(f):
+    if isinstance(f.newlines, six.text_type):
+        return f.newlines
+
+    return DEFAULT_NEWLINES
 
 
 if PIPENV_PIPFILE:
@@ -72,7 +88,7 @@ if PIPENV_TEST_INDEX:
     }
 else:
     DEFAULT_SOURCE = {
-        u'url': u'https://pypi.python.org/simple',
+        u'url': u'https://pypi.org/simple',
         u'verify_ssl': True,
         u'name': u'pypi',
     }
@@ -92,8 +108,10 @@ class Project(object):
         self._name = None
         self._virtualenv_location = None
         self._download_location = None
-        self._proper_names_location = None
+        self._proper_names_db_path = None
         self._pipfile_location = None
+        self._pipfile_newlines = DEFAULT_NEWLINES
+        self._lockfile_newlines = DEFAULT_NEWLINES
         self._requirements_location = None
         self._original_dir = os.path.abspath(os.curdir)
         self.which = which
@@ -201,6 +219,12 @@ class Project(object):
     def requirements_exists(self):
         return bool(self.requirements_location)
 
+    def is_venv_in_project(self):
+        return PIPENV_VENV_IN_PROJECT or (
+            self.project_directory and
+            os.path.exists(os.path.join(self.project_directory, '.venv'))
+        )
+
     @property
     def virtualenv_exists(self):
         # TODO: Decouple project from existence of Pipfile.
@@ -217,7 +241,6 @@ class Project(object):
 
     @classmethod
     def _get_virtualenv_location(cls, name):
-        from pipenv.patched.pew.pew import get_workon_home
         venv = get_workon_home() / name
         if not venv.exists():
             return ''
@@ -256,24 +279,24 @@ class Project(object):
         # This should work most of the time, for non-WIndows, in-project venv,
         # or "proper" path casing (on Windows).
         if (os.name != 'nt' or
-                PIPENV_VENV_IN_PROJECT or
+                self.is_venv_in_project() or
                 self._get_virtualenv_location(venv_name)):
             return clean_name, encoded_hash
 
         # Check for different capitalization of the same project.
-        from pipenv.patched.pew.pew import lsenvs
+        from .patched.pew.pew import lsenvs
         for env in lsenvs():
-            env_name = env[:-9]
-            if not (env[-9] != '-' and
-                    env[-8:].isalpha() and
-                    env_name.lower() != name.lower()):
+            try:
+                env_name, hash_ = env.rsplit('-', 1)
+            except ValueError:
+                continue
+            if len(hash_) != 8 or env_name.lower() != name.lower():
                 continue
             return get_name(env_name, self.pipfile_location.replace(name, env_name))
 
         # Use the default if no matching env exists.
         return clean_name, encoded_hash
 
-
     @property
     def virtualenv_name(self):
         sanitized, encoded_hash = self._get_virtualenv_hash(self.name)
@@ -291,11 +314,9 @@ class Project(object):
         # Use cached version, if available.
         if self._virtualenv_location:
             return self._virtualenv_location
-        venv_in_project = PIPENV_VENV_IN_PROJECT or \
-            os.path.exists(os.path.join(self.project_directory, '.venv'))
 
         # Default mode.
-        if not venv_in_project:
+        if not self.is_venv_in_project():
             loc = self._get_virtualenv_location(self.virtualenv_name)
         # The user wants the virtualenv in the project.
         else:
@@ -321,25 +342,24 @@ class Project(object):
         return self._download_location
 
     @property
-    def proper_names_location(self):
-        if self._proper_names_location is None:
-            loc = os.sep.join(
-                [self.virtualenv_location, 'pipenv-proper-names.txt']
+    def proper_names_db_path(self):
+        if self._proper_names_db_path is None:
+            self._proper_names_db_path = Path(
+                self.virtualenv_location,
+                'pipenv-proper-names.txt',
             )
-            self._proper_names_location = loc
-        # Create the database, if it doesn't exist.
-        open(self._proper_names_location, 'a').close()
-        return self._proper_names_location
+        self._proper_names_db_path.touch()  # Ensure the file exists.
+        return self._proper_names_db_path
 
     @property
     def proper_names(self):
-        with open(self.proper_names_location) as f:
+        with self.proper_names_db_path.open() as f:
             return f.read().splitlines()
 
     def register_proper_name(self, name):
         """Registers a proper name to the database."""
-        with open(self.proper_names_location, 'a') as f:
-            f.write('{0}\n'.format(name))
+        with self.proper_names_db_path.open('a') as f:
+            f.write(u'{0}\n'.format(name))
 
     @property
     def pipfile_location(self):
@@ -369,9 +389,7 @@ class Project(object):
         """Parse Pipfile into a TOMLFile and cache it
 
         (call clear_pipfile_cache() afterwards if mutating)"""
-        # Open the pipfile, read it into memory.
-        with open(self.pipfile_location) as f:
-            contents = f.read()
+        contents = self.read_pipfile()
         # use full contents to get around str/bytes 2/3 issues
         cache_key = (self.pipfile_location, contents)
         if cache_key not in _pipfile_cache:
@@ -379,10 +397,17 @@ class Project(object):
             _pipfile_cache[cache_key] = parsed
         return _pipfile_cache[cache_key]
 
+    def read_pipfile(self):
+        # Open the pipfile, read it into memory.
+        with io.open(self.pipfile_location) as f:
+            contents = f.read()
+            self._pipfile_newlines = preferred_newlines(f)
+
+        return contents
+
     @property
     def pased_pure_pipfile(self):
-        with open(self.pipfile_location) as f:
-            contents = f.read()
+        contents = self.read_pipfile()
 
         return self._parse_pipfile(contents)
 
@@ -474,50 +499,43 @@ class Project(object):
 
     @property
     def lockfile_content(self):
-        with open(self.lockfile_location) as lock:
-            j = json.load(lock)
+        return self.load_lockfile()
 
-        # Expand environment variables in Pipfile.lock at runtime.
-        for i, source in enumerate(j['_meta']['sources'][:]):
-            j['_meta']['sources'][i]['url'] = os.path.expandvars(j['_meta']['sources'][i]['url'])
+    def _get_editable_packages(self, dev=False):
+        section = 'dev-packages' if dev else 'packages'
+        packages = {
+            k: v
+            for k, v in self.parsed_pipfile.get(section, {}).items()
+            if is_editable(v)
+        }
+        return packages
 
-        return j
+    def _get_vcs_packages(self, dev=False):
+        section = 'dev-packages' if dev else 'packages'
+        packages = {
+            k: v
+            for k, v in self.parsed_pipfile.get(section, {}).items()
+            if is_vcs(v) or is_vcs(k)
+        }
+        return packages or {}
 
     @property
     def editable_packages(self):
-        packages = {
-            k: v
-            for k, v in self.parsed_pipfile.get('packages', {}).items()
-            if is_editable(v)
-        }
-        return packages
+        return self._get_editable_packages(dev=False)
 
     @property
     def editable_dev_packages(self):
-        packages = {
-            k: v
-            for k, v in self.parsed_pipfile.get('dev-packages', {}).items()
-            if is_editable(v)
-        }
-        return packages
+        return self._get_editable_packages(dev=True)
 
     @property
     def vcs_packages(self):
         """Returns a list of VCS packages, for not pip-tools to consume."""
-        ps = {}
-        for k, v in self.parsed_pipfile.get('packages', {}).items():
-            if is_vcs(v) or is_vcs(k):
-                ps.update({k: v})
-        return ps
+        return self._get_vcs_packages(dev=False)
 
     @property
     def vcs_dev_packages(self):
         """Returns a list of VCS packages, for not pip-tools to consume."""
-        ps = {}
-        for k, v in self.parsed_pipfile.get('dev-packages', {}).items():
-            if is_vcs(v) or is_vcs(k):
-                ps.update({k: v})
-        return ps
+        return self._get_vcs_packages(dev=True)
 
     @property
     def all_packages(self):
@@ -546,17 +564,19 @@ class Project(object):
         if not self.pipfile_exists:
             return True
 
-        with open(self.pipfile_location, 'r') as f:
-            if not f.read():
-                return True
+        if not len(self.read_pipfile()):
+            return True
 
         return False
 
     def create_pipfile(self, python=None):
         """Creates the Pipfile, filled with juicy defaults."""
+        from .patched.notpip._internal import ConfigOptionParser
+        from .patched.notpip._internal.cmdoptions import make_option_group, index_group
         config_parser = ConfigOptionParser(name=self.name)
-        install = dict(config_parser.get_config_section('install'))
-        indexes = install.get('extra-index-url', '').lstrip('\n').split('\n')
+        config_parser.add_option_group(make_option_group(index_group, config_parser))
+        install = config_parser.option_groups[0]
+        indexes = ' '.join(install.get_option('--extra-index-url').default).lstrip('\n').split('\n')
         sources = [DEFAULT_SOURCE]
         for i, index in enumerate(indexes):
             if not index:
@@ -571,6 +591,7 @@ class Project(object):
                     u'name': source_name,
                 }
             )
+
         data = {
             u'source': sources,
             # Default packages.
@@ -578,12 +599,17 @@ class Project(object):
             u'dev-packages': {},
         }
         # Default requires.
-        required_python = python or self.which(
-            'python', self.virtualenv_location
-        )
-        data[u'requires'] = {
-            'python_version': python_version(required_python)[: len('2.7')]
-        }
+        required_python = python
+        if not python:
+            if self.virtualenv_location:
+                required_python = self.which('python', self.virtualenv_location)
+            else:
+                required_python = self.which('python')
+        version = python_version(required_python) or PIPENV_DEFAULT_PYTHON_VERSION
+        if version and len(version) >= 3:
+            data[u'requires'] = {
+                'python_version': version[: len('2.7')]
+            }
         self.write_toml(data, 'Pipfile')
 
     def write_toml(self, data, path=None):
@@ -594,7 +620,7 @@ class Project(object):
             formatted_data = contoml.dumps(data).rstrip()
         except Exception:
             for section in ('packages', 'dev-packages'):
-                for package in data[section]:
+                for package in data.get(section, {}):
                     # Convert things to inline tables — fancy :)
                     if hasattr(data[section][package], 'keys'):
                         _data = data[section][package]
@@ -603,44 +629,84 @@ class Project(object):
                         )
                         data[section][package].update(_data)
             formatted_data = toml.dumps(data).rstrip()
+
+        if Path(path).absolute() == Path(self.pipfile_location).absolute():
+            newlines = self._pipfile_newlines
+        else:
+            newlines = DEFAULT_NEWLINES
         formatted_data = cleanup_toml(formatted_data)
-        with open(path, 'w') as f:
+        with io.open(path, 'w', newline=newlines) as f:
             f.write(formatted_data)
         # pipfile is mutated!
         self.clear_pipfile_cache()
 
+    def write_lockfile(self, content):
+        """Write out the lockfile.
+        """
+        newlines = self._lockfile_newlines
+        s = simplejson.dumps(   # Send Unicode in to guarentee Unicode out.
+            content, indent=4, separators=(u',', u': '), sort_keys=True,
+        )
+        with atomic_open_for_write(self.lockfile_location, newline=newlines) as f:
+            f.write(s)
+            if not s.endswith(u'\n'):
+                f.write(u'\n')  # Write newline at end of document. GH #319.
+
+    @property
+    def pipfile_sources(self):
+        if 'source' not in self.parsed_pipfile:
+            return [DEFAULT_SOURCE]
+        # We need to make copies of the source info so we don't
+        # accidentally modify the cache. See #2100 where values are
+        # written after the os.path.expandvars() call.
+        return [
+            {k: safe_expandvars(v) for k, v in source.items()}
+            for source in self.parsed_pipfile['source']
+        ]
+
     @property
     def sources(self):
-        if self.lockfile_exists:
+        if self.lockfile_exists and hasattr(self.lockfile_content, 'keys'):
             meta_ = self.lockfile_content['_meta']
             sources_ = meta_.get('sources')
             if sources_:
                 return sources_
 
-        if 'source' in self.parsed_pipfile:
-            return self.parsed_pipfile['source']
         else:
-            return [DEFAULT_SOURCE]
+            return self.pipfile_sources
+
+    def find_source(self, source):
+        """given a source, find it.
+
+        source can be a url or an index name.
+        """
+        if not is_valid_url(source):
+            try:
+                source = self.get_source(name=source)
+            except SourceNotFound:
+                source = self.get_source(url=source)
+        else:
+            source = self.get_source(url=source)
+        return source
 
     def get_source(self, name=None, url=None):
-        for source in self.sources:
+        def find_source(sources, name=None, url=None):
+            source = None
             if name:
-                if source.get('name') == name:
-                    return source
-
+                source = [s for s in sources if s.get('name') == name]
             elif url:
-                if source.get('url') in url:
-                    return source
+                source = [s for s in sources if url.startswith(s.get('url'))]
+            if source:
+                return first(source)
+
+        found_source = find_source(self.sources, name=name, url=url)
+        if found_source:
+            return found_source
+        found_source = find_source(self.pipfile_sources, name=name, url=url)
+        if found_source:
+            return found_source
         raise SourceNotFound(name or url)
 
-    def destroy_lockfile(self):
-        """Deletes the lockfile."""
-        try:
-            return os.remove(self.lockfile_location)
-
-        except OSError:
-            pass
-
     def get_package_name_in_pipfile(self, package_name, dev=False):
         """Get the equivalent package name in pipfile"""
         key = 'dev-packages' if dev else 'packages'
@@ -661,27 +727,22 @@ class Project(object):
             self.write_toml(p)
 
     def add_package_to_pipfile(self, package_name, dev=False):
+        from .vendor.requirementslib import Requirement
         # Read and append Pipfile.
         p = self.parsed_pipfile
         # Don't re-capitalize file URLs or VCSs.
-        converted = convert_deps_from_pip(package_name)
-        converted = converted[[k for k in converted.keys()][0]]
-        if not (
-            is_file(package_name) or is_vcs(converted) or 'path' in converted
-        ):
-            package_name = pep423_name(package_name)
+        package = Requirement.from_line(package_name.strip())
+        _, converted = package.pipfile_entry
         key = 'dev-packages' if dev else 'packages'
         # Set empty group if it doesn't exist yet.
         if key not in p:
             p[key] = {}
-        package = convert_deps_from_pip(package_name)
-        package_name = [k for k in package.keys()][0]
-        name = self.get_package_name_in_pipfile(package_name, dev)
-        if name and converted == '*':
+        name = self.get_package_name_in_pipfile(package.name, dev)
+        if name and is_star(converted):
             # Skip for wildcard version
             return
         # Add the package to the group.
-        p[key][name or package_name] = package[package_name]
+        p[key][name or package.normalized_name] = converted
         # Write Pipfile.
         self.write_toml(p)
 
@@ -702,13 +763,30 @@ class Project(object):
         if self.ensure_proper_casing():
             self.write_toml(self.parsed_pipfile)
 
+    def load_lockfile(self, expand_env_vars=True):
+        with io.open(self.lockfile_location) as lock:
+            j = json.load(lock)
+            self._lockfile_newlines = preferred_newlines(lock)
+        # lockfile is just a string
+        if not j or not hasattr(j, 'keys'):
+            return j
+
+        if expand_env_vars:
+            # Expand environment variables in Pipfile.lock at runtime.
+            for i, source in enumerate(j['_meta']['sources'][:]):
+                j['_meta']['sources'][i]['url'] = os.path.expandvars(j['_meta']['sources'][i]['url'])
+
+        return j
+
     def get_lockfile_hash(self):
         if not os.path.exists(self.lockfile_location):
             return
-        # Open the lockfile.
-        with codecs.open(self.lockfile_location, 'r') as f:
-            lockfile = json.load(f)
-        return lockfile['_meta'].get('hash', {}).get('sha256')
+
+        lockfile = self.load_lockfile(expand_env_vars=False)
+        if '_meta' in lockfile and hasattr(lockfile, 'keys'):
+            return lockfile['_meta'].get('hash', {}).get('sha256')
+        # Lockfile exists but has no hash at all
+        return ''
 
     def calculate_pipfile_hash(self):
         # Update the lockfile if it is out-of-date.
